from datetime import datetime
import time
from typing import Any, Optional, Callable
import jwt
import requests

# These provide AWS cognito authentication support
from pycognito import Cognito

CLIENT_ID = '4qte47jbstod8apnfic0bunmrq'
USER_POOL = 'us-east-2_ghlOXVLi1'

class Auth:
    def __init__(
        self,
        host: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        connect_timeout: float = 6.03,
        read_timeout: float = 10.03,
        tokens: Optional['dict[str, Any]'] = None,
        token_updater: Optional[Callable[['dict[str, Any]'], None]] = None,
        max_retry_attempts: int = 5,
        initial_retry_delay: float = 0.5,
        max_retry_delay: float = 30.0,
    ):
        self.host = host
        self.connect_timeout = connect_timeout
        self.read_timeout = read_timeout
        self.token_updater = token_updater
        self.max_retry_attempts = max(max_retry_attempts, 1)
        self.initial_retry_delay = max(initial_retry_delay, 0)
        self.max_retry_delay = max(max_retry_delay, 0)
        self.pool_wellknown_jwks = None

        if tokens and tokens['access_token'] and tokens['id_token'] and tokens['refresh_token']:
            # use existing tokens
            self.cognito = Cognito(USER_POOL, CLIENT_ID,
                user_pool_region='us-east-2', 
                id_token=tokens['id_token'], 
                access_token=tokens['access_token'], 
                refresh_token=tokens['refresh_token'])
        elif username and password:
            #log in with username and password
            self.cognito = Cognito(USER_POOL, CLIENT_ID, 
                user_pool_region='us-east-2', username=username)
            self.cognito.authenticate(password=password)

        self.tokens = self.refresh_tokens()

    def refresh_tokens(self) -> 'dict[str, str]':
        """Refresh and return new tokens."""
        self.cognito.renew_access_token()
        tokens = self._extract_tokens_from_cognito()

        if self.token_updater is not None:
            self.token_updater(tokens)

        return tokens

    def get_username(self) -> str:
        """Get the username associated with the logged in user."""
        user = self.cognito.get_user()
        return user._data['email']

    def request(self, method: str, path: str, **kwargs) -> requests.Response:
        """Make a request."""
        dec_access_token = self._decode_token(self.tokens['access_token'])

        attempts = 0
        while attempts < self.max_retry_attempts:
            attempts += 1
            if datetime.now() > datetime.fromtimestamp(dec_access_token["exp"]):
                # expired, get new tokens
                self.tokens = self.refresh_tokens()

            response = self._do_request(method, path, **kwargs)

            if response.status_code == 401:
                # if unauthorized, try refreshing the tokens
                self.tokens = self.refresh_tokens()
                # then run the request again with updated tokens
                response = self._do_request(method, path, **kwargs)
            
            if response.status_code >= 500:
                # if server error, retry with exponential backoff
                delay = min(
                    self.initial_retry_delay * (2 ** (attempts - 1)),
                    self.max_retry_delay,
                )
                print(f"Server error {response.status_code}, retrying in {delay} seconds")
                time.sleep(delay)
                continue

            if response.status_code < 500:
                return response

        return response

    def _extract_tokens_from_cognito(self) -> 'dict[str, Any]':
        return {
            'access_token': self.cognito.access_token,
            'id_token': self.cognito.id_token, # Emporia uses this token for authentication
            'refresh_token': self.cognito.refresh_token,
            'token_type': self.cognito.token_type
        }

    def _do_request(self, method: str, path: str, **kwargs) -> requests.Response:
        headers = kwargs.get("headers")

        if headers is None:
            headers = {}
        else:
            headers = dict(headers)
        headers["authtoken"] = self.tokens['id_token']

        return requests.request(
            method, f"{self.host}/{path}", **kwargs, headers=headers,
            timeout=(self.connect_timeout, self.read_timeout),
        )
    
    def _decode_token(self, token: str) -> dict:
        '''Decode a JWT token and return the payload as a dictionary, without a hard dependency on pycognito.'''
        if not self.pool_wellknown_jwks:
            self.pool_wellknown_jwks = requests.get(
                f"https://cognito-idp.us-east-2.amazonaws.com/{USER_POOL}/.well-known/jwks.json", timeout=5
            ).json()
        
        kid = jwt.get_unverified_header(token).get("kid")
        keys = self.pool_wellknown_jwks.get("keys")
        key = list(filter(lambda x: x.get("kid") == kid, keys))[0]
        hmac_key = jwt.api_jwk.PyJWK(key).key
        return jwt.api_jwt.decode(token, algorithms=["RS256"], key=hmac_key, options={"verify_exp": False, "verify_iat": False, "verify_nbf": False})


class SimulatedAuth(Auth):
    def __init__(self, host: str, username: Optional[str] = None, password: Optional[str] = None):
        self.host = host
        self.username = username
        self.password = password
        self.connect_timeout = 6.03
        self.read_timeout = 10.03

        self.tokens = self.refresh_tokens()
    
    def refresh_tokens(self) -> dict[str, str]:
        return {'id_token': 'simulator'}

    def get_username(self) -> str:
        """Get the username associated with the logged in user."""
        return self.username or "simulator"

    def request(self, method: str, path: str, **kwargs) -> requests.Response:
        """Make a request."""
        response = self._do_request(method, path, **kwargs)

        if response.status_code == 401:
            # if unauthorized, try refreshing the tokens
            self.tokens = self.refresh_tokens()
            # then run the request again with updated tokens
            response = self._do_request(method, path, **kwargs)

        return response


from typing import TYPE_CHECKING, Optional
from pydantic import BaseModel
import datetime


def _format_time(time: datetime.datetime) -> str:
    """Convert time to utc, then format"""
    # check if aware
    if time.tzinfo and time.tzinfo.utcoffset(time) is not None:
        # aware, convert to utc
        time = time.astimezone(datetime.timezone.utc)
    else:
        # unaware, assume it's already utc
        time = time.replace(tzinfo=datetime.timezone.utc)
    time = time.replace(tzinfo=None)  # make it unaware
    iso_formatted = time.isoformat()
    # drop the milliseconds
    return iso_formatted[: iso_formatted.rfind(".")] + "Z"


class SimulatorBase(BaseModel):
    class Config:
        json_encoders = {datetime.datetime: _format_time}


class SimulatorCustomer(SimulatorBase):
    customerGid: int
    email: str
    firstName: str
    lastName: str
    createdAt: datetime.datetime


class SimulatorOutlet(SimulatorBase):
    deviceGid: int
    outletOn: bool
    loadGid: int


class SimulatorChargerRequest(SimulatorBase):
    deviceGid: int
    loadGid: int
    chargerOn: bool
    chargingRate: int
    maxChargingRate: int
    breakerPIN: Optional[str] = None


class SimulatorCharger(SimulatorBase):
    deviceGid: int
    loadGid: int
    message: str = "EV is not accepting charge"
    status: str = "Standby"
    icon: str = "CarConnected"
    iconLabel: str = "Offering Charge"
    debugCode: str = "311"
    iconDetailText: str = "Check your vehicle for a scheduled charge time."
    faultText: Optional[str] = None
    proControlCode: Optional[str] = None
    breakerPIN: Optional[str] = "0000"
    chargerOn: bool = True
    chargingRate: int = 40
    maxChargingRate: int = 40
    loadManagementEnabled: bool = False
    hideChargeRateSliderText: Optional[str] = None

    def set_to_standby(self):
        self.message = "EV is not accepting charge"
        self.status = "Standby"
        self.icon = "CarConnected"
        self.iconLabel = "Offering Charge"
        self.debugCode = "311"
        self.iconDetailText = "Check your vehicle for a scheduled charge time."


class SimulatorDeviceConnected(SimulatorBase):
    deviceGid: int
    connected: bool
    offlineSince: Optional[datetime.datetime]


class SimulatorChannel(SimulatorBase):
    deviceGid: int
    name: Optional[str]
    channelNum: str
    channelMultiplier: float = 1.0
    channelTypeGid: Optional[int]


class SimulatorLatitudeLongitude(SimulatorBase):
    latitude: float
    longitude: float


class SimulatorLocationInformation(SimulatorBase):
    airConditioning: str = "true"
    heatSource: str = "electricFurnace"
    locationSqFt: str = "2000"
    numElectricCars: str = "0"
    locationType: str = "houseMultiLevel"
    numPeople: str = "4"
    swimmingPool: str = "false"
    hotTub: str = "false"
    primaryVehicle: Optional[str] = None


class SimulatorLocationProperties(SimulatorBase):
    timeZone: str
    latitudeLongitude: SimulatorLatitudeLongitude
    utilityRateGid: Optional[int]
    deviceName: Optional[str]
    deviceGid: int
    zipCode: str
    billingCycleStartDay: int
    usageCentPerKwHour: float
    peakDemandDollarPerKw: float
    locationInformation: SimulatorLocationInformation


class SimulatorSubDevice(SimulatorBase):
    deviceGid: int
    manufacturerDeviceId: str
    model: str
    firmware: Optional[str]
    channels: list[SimulatorChannel]


class SimulatorDevice(SimulatorBase):
    deviceGid: int
    manufacturerDeviceId: str
    model: str
    firmware: Optional[str]
    parentDeviceGid: Optional[int]
    parentChannelNum: Optional[str]
    locationProperties: SimulatorLocationProperties
    outlet: Optional[SimulatorOutlet]
    evCharger: Optional[SimulatorCharger]
    battery: Optional[dict]
    deviceConnected: SimulatorDeviceConnected
    devices: list[SimulatorSubDevice]
    channels: list[SimulatorChannel]


class CustomerDevicesResponse(SimulatorBase):
    customerGid: int
    email: str
    firstName: str
    lastName: str
    createdAt: datetime.datetime
    devices: list[SimulatorDevice]


class ChannelType(SimulatorBase):
    channelTypeGid: int
    description: str
    selectable: bool


class SimulatorLoad(SimulatorBase):
    loadGid: int
    schedulesEnabled: bool
    nextScheduledEventText: Optional[str]
    peakDemandEnabled: bool
    excessGenerationEnabled: bool
    energyManagementText: Optional[str]
    energyManagementOverridden: bool
    warningText: Optional[str]


class StatusResponse(SimulatorBase):
    devicesConnected: list[SimulatorDeviceConnected]
    evChargers: list[SimulatorCharger]
    loads: list[SimulatorLoad]
    batteries: list[dict]
    outlets: list[SimulatorOutlet]


class CreateOutletRequest(SimulatorBase):
    deviceGid: int
    outletOn: bool = True
    name: Optional[str] = None
    parentDeviceGid: Optional[int] = None
    parentChannelNum: Optional[str] = None


class CreateChargerRequest(SimulatorBase):
    deviceGid: int
    chargerOn: bool = True
    name: Optional[str] = None
    breakerSize: int = 50
    parentDeviceGid: Optional[int] = None
    parentChannelNum: Optional[str] = None


class CreateVueRequest(SimulatorBase):
    deviceGid: int
    name: Optional[str] = None
    channelCount: int = 8
    parentDeviceGid: Optional[int] = None
    parentChannelNum: Optional[str] = None

if TYPE_CHECKING:
    from .models import DeviceUsage  # Avoid circular import

class ChannelUsage(SimulatorBase):
    name: str
    percentage: float
    nestedDevices: list['DeviceUsage']
    usage: Optional[float]
    deviceGid: int
    channelNum: str

class DeviceUsage(SimulatorBase):
    deviceGid: int
    channelUsages: list[ChannelUsage]

class DeviceListUsage(SimulatorBase):
    devices: list[DeviceUsage]
    energyUnit: str
    instant: datetime.datetime
    scale: str

class DeviceUsageResponse(SimulatorBase):
    deviceListUsages: DeviceListUsage


import pyemvue
import json
import getpass
import os

# Create PyEmVue object
vue = pyemvue.PyEmVue()

# Check if keys.json exists
if not os.path.exists('keys.json'):
    # Get email and password
    email = input("Enter your email: ")
    password = getpass.getpass("Enter your password: ")

    # Log in to PyEmVue
    vue.login(username=email, password=password, token_storage_file='keys.json')
else:
    # If keys.json exists, log in using the keys in the file
    vue.login(token_storage_file='keys.json')

# Get the list of chargers
chargers = vue.get_chargers()

# Print charger data to the screen
print("Charger data:")
print(json.dumps([charger.__dict__ for charger in chargers], indent=4))

# Wait for user to press enter
input("Press enter to turn off the charger and print the charger data again...")

# Turn off charger and print the charger data again
for charger in chargers:
    vue.update_charger(charger, on=False)
    print("Updated charger data:")
    print(json.dumps(charger.__dict__, indent=4))


from fastapi import FastAPI, Request, Response
from fastapi.responses import JSONResponse

from simulator.models import *
from simulator.simulator_state import SimulatorState

state = SimulatorState()

# Set up a default home with an 8 channel Vue, 4 outlets, and 1 EV charger
state.add_vue(1000, "Home", channelCount=8)
state.add_outlet(1001, "plug1", True, parentDeviceGid=1000, parentChannelNum="1")
state.add_outlet(1002, "plug2", False, parentDeviceGid=1000, parentChannelNum="1")
state.add_outlet(1003, "plug3", True, parentDeviceGid=1000, parentChannelNum="4")
state.add_outlet(1004, "plug4", False, parentDeviceGid=1000, parentChannelNum="1,2,3")
state.add_charger(
    1005, "EV", True, breakerSize=50, parentDeviceGid=1000, parentChannelNum="1,2,3"
)
# the EV charger is pulling 40 amps at 240 volts
state.set_channel_1min_watts(1005, "1,2,3", 40 * 240)

# plug 3 is pulling 10 amps at 120 volts
state.set_channel_1min_watts(1003, "1,2,3", 10 * 120)

# plug 1 is pulling 5 amps at 120 volts
state.set_channel_1min_watts(1001, "1,2,3", 5 * 120)

# the overall house is pulling 95 amps at 240 volts
state.set_channel_1min_watts(1000, "1,2,3", 95 * 240)

# the balance is effectively 40 amps
state.set_channel_1min_watts(1000, "Balance", 40 * 240)


app = FastAPI()

# All APIs that are supported are:
# API_CHANNELS = 'devices/{deviceGid}/channels'
# API_CHANNEL_TYPES = 'devices/channels/channeltypes'
# API_CHARGER = 'devices/evcharger'
# API_CHART_USAGE = 'AppAPI?apiMethod=getChartUsage&deviceGid={deviceGid}&channel={channel}&start={start}&end={end}&scale={scale}&energyUnit={unit}'
# API_CUSTOMER = 'customers'
# API_CUSTOMER_DEVICES = 'customers/devices'
# API_DEVICES_USAGE = 'AppAPI?apiMethod=getDeviceListUsages&deviceGids={deviceGids}&instant={instant}&scale={scale}&energyUnit={unit}'
# API_DEVICE_PROPERTIES = 'devices/{deviceGid}/locationProperties'
# API_GET_STATUS = 'customers/devices/status'
# API_OUTLET = 'devices/outlet'
# API_VEHICLES = 'customers/vehicles'
# API_VEHICLE_STATUS = 'vehicles/v2/settings?vehicleGid={vehicleGid}'


class CustomException(Exception):
    def __init__(self, status_code: int, message: str):
        self.status_code = status_code
        self.message = message


class NotAuthorizedException(CustomException):
    def __init__(self, deviceGid: int):
        self.status_code = 401
        self.message = f"{state.customer.email} is not authorized on the requested deviceGid {deviceGid}"


@app.exception_handler(CustomException)
async def unicorn_exception_handler(request: Request, exc: CustomException):
    return JSONResponse(
        status_code=exc.status_code,
        content={"message": exc.message},
    )


@app.get("/customers")
def get_customers() -> SimulatorCustomer:
    return state.customer


@app.get("/customers/devices")
def get_customers_devices() -> CustomerDevicesResponse:
    return state.get_customers_devices()


@app.get("/devices/{deviceGid}/locationProperties")
def get_devices_locationProperties(deviceGid: int) -> SimulatorLocationProperties:
    for device in state.devices:
        if device.deviceGid == deviceGid:
            return device.locationProperties
    # return a 401 if the deviceGid is not found
    raise NotAuthorizedException(deviceGid)


@app.get("/devices/channels/channeltypes")
def get_devices_channels_channelTypes() -> list[ChannelType]:
    return state.channel_types


@app.get("/customers/devices/status")
def get_customers_devices_status() -> StatusResponse:
    return state.get_status()


@app.put("/devices/outlet")
def put_devices_outlet(outlet: SimulatorOutlet) -> SimulatorOutlet:
    for existingOutlet in state.outlets:
        if existingOutlet.deviceGid == outlet.deviceGid:
            existingOutlet.outletOn = outlet.outletOn
            existingOutlet.loadGid = outlet.loadGid
            return existingOutlet
    # return a 401 if the deviceGid is not found
    raise NotAuthorizedException(outlet.deviceGid)


@app.put("/devices/evcharger")
def put_devices_evcharger(charger: SimulatorChargerRequest) -> SimulatorCharger:
    for existingCharger in state.chargers:
        if existingCharger.deviceGid == charger.deviceGid:
            existingCharger.chargerOn = charger.chargerOn
            existingCharger.chargingRate = charger.chargingRate
            existingCharger.maxChargingRate = charger.maxChargingRate
            existingCharger.breakerPIN = charger.breakerPIN
            return existingCharger
    # return a 401 if the deviceGid is not found
    raise NotAuthorizedException(charger.deviceGid)


# API_DEVICES_USAGE = 'AppAPI?apiMethod=getDeviceListUsages&deviceGids={deviceGids}&instant={instant}&scale={scale}&energyUnit={unit}'
@app.get("/AppAPI")
def get_app_api(
    deviceGids: Optional[str] = None,
    instant: Optional[datetime.datetime] = None,
    scale: str = "1MIN",
    energyUnit: str = "KilowattHours",
) -> DeviceUsageResponse:
    return state.get_devices_usage(
        deviceGids, instant or datetime.datetime.utcnow(), scale, energyUnit
    )


# meta APIs for controlling the simulator
@app.post("/simulator/vue")
def post_create_vue(vue: CreateVueRequest) -> SimulatorDevice:
    created = state.add_vue(
        vue.deviceGid,
        vue.name,
        vue.channelCount,
        vue.parentDeviceGid,
        vue.parentChannelNum,
    )
    return created


@app.post("/simulator/outlet")
def post_create_outlet(outlet: CreateOutletRequest) -> SimulatorOutlet:
    created = state.add_outlet(
        outlet.deviceGid,
        outlet.name,
        outlet.outletOn,
        outlet.parentDeviceGid,
        outlet.parentChannelNum,
    )
    return created


@app.post("/simulator/charger")
def post_create_charger(charger: CreateChargerRequest) -> SimulatorCharger:
    created = state.add_charger(
        charger.deviceGid,
        charger.name,
        charger.chargerOn,
        charger.breakerSize,
        charger.parentDeviceGid,
        charger.parentChannelNum,
    )
    return created


@app.delete("/simulator/device/{deviceGid}")
def delete_device(deviceGid: int, response: Response) -> SimulatorDevice:
    deleted = state.delete_device(deviceGid)
    if deleted is None:
        response.status_code = 404
        return response  # type: ignore
    return deleted


@app.put("/simulator/device/{deviceGid}/channel/{channelNum}/usage")
def put_channel_usage(
    deviceGid: int,
    channelNum: str,
    watts: Optional[float] = None,
    usage: Optional[float] = None,
    scale: Optional[str] = "1MIN",
) -> None:
    if scale == "1MIN":
        if usage is not None:  # usage takes precedence
            state.set_channel_1min_usage(deviceGid, channelNum, usage)
        else:
            state.set_channel_1min_watts(deviceGid, channelNum, watts)
        # probably should return the state of the channel here
        return None
    # other scales are not supported yet



import datetime
from typing import Any, Optional
from typing_extensions import Self
from dateutil.parser import parse

class VueDevice(object):
    def __init__(self, gid=0, manId='', modelNum='', firmwareVersion=''):
        self.device_gid: int = gid
        self.manufacturer_id = manId
        self.model = modelNum
        self.firmware = firmwareVersion
        self.parent_device_gid: int = 0
        self.parent_channel_num: str = ''
        self.channels: list[VueDeviceChannel] = []
        self.outlet: Optional[OutletDevice] = None
        self.ev_charger: Optional[ChargerDevice] = None

        self.connected: bool = False
        self.offline_since = datetime.datetime.min

        #extra info
        self.device_name = ''
        self.zip_code = '00000'
        self.time_zone = ''
        self.usage_cent_per_kw_hour = 0.0
        self.peak_demand_dollar_per_kw = 0.0
        self.billing_cycle_start_day = 0
        self.solar = False
        self.air_conditioning = 'false'
        self.heat_source = ''
        self.location_sqft = '0'
        self.num_electric_cars = '0'
        self.location_type = ''
        self.num_people = ''
        self.swimming_pool = 'false'
        self.hot_tub = 'false'
        self.latitude = 0
        self.longitude = 0
        self.utility_rate_gid = None


    def from_json_dictionary(self, js: 'dict[str, Any]') -> Self:
        """Populate device data from a dictionary extracted from the response json."""
        if 'deviceGid' in js: self.device_gid = js['deviceGid']
        if 'manufacturerDeviceId' in js: self.manufacturer_id = js['manufacturerDeviceId']
        if 'model' in js: self.model = js['model']
        if 'firmware' in js: self.firmware = js['firmware']
        if 'parentDeviceGid' in js: self.parent_device_gid = js['parentDeviceGid']
        if 'parentChannelNum' in js: self.parent_channel_num = js['parentChannelNum']
        if 'locationProperties' in js:
            self.populate_location_properties_from_json(js['locationProperties'])
        # 'devices' is empty in my system, will add support later if possible
        if 'channels' in js:
            # Channels are another subtype and the channelNum is used in other calls
            self.channels = []
            for chnl in js['channels']:
                self.channels.append(VueDeviceChannel().from_json_dictionary(chnl))
        # outlets are a special type
        if 'outlet' in js and js['outlet']:
            self.outlet = OutletDevice().from_json_dictionary(js['outlet'])
        # EVSEs are also special
        if 'evCharger' in js and js['evCharger']:
            self.ev_charger = ChargerDevice().from_json_dictionary(js['evCharger'])

        # Online data
        if 'deviceConnected' in js and js['deviceConnected']:
            con = js['deviceConnected']
            if 'connected' in con: self.connected = con['connected']
            try:
                if 'offlineSince' in con and con['offlineSince']: self.offline_since = parse(con['offlineSince'])
            except:
                self.offline_since = datetime.datetime.min
        return self
    
    def populate_location_properties_from_json(self, js: 'dict[str, Any]'):
        """Adds the values from the get_device_properties method."""
        if 'deviceName' in js: self.device_name = js['deviceName']
        if 'zipCode' in js: self.zip_code = js['zipCode']
        if 'timeZone' in js: self.time_zone = js['timeZone']
        if 'usageCentPerKwHour' in js: self.usage_cent_per_kw_hour = js['usageCentPerKwHour']
        if 'peakDemandDollarPerKw' in js: self.peak_demand_dollar_per_kw = js['peakDemandDollarPerKw']
        if 'billingCycleStartDay' in js: self.billing_cycle_start_day = js['billingCycleStartDay']
        if 'solar' in js: self.solar = js['solar']
        if 'utilityRateGid' in js: self.utility_rate_gid = js['utilityRateGid']
        if 'locationInformation' in js and js['locationInformation']:
            li = js['locationInformation']
            if 'airConditioning' in li: self.air_conditioning = li['airConditioning']
            if 'heatSource' in li: self.heat_source = li['heatSource']
            if 'locationSqFt' in li: self.location_sqft = li['locationSqFt']
            if 'numElectricCars' in li: self.num_electric_cars = li['numElectricCars']
            if 'locationType' in li: self.location_type = li['locationType']
            if 'numPeople' in li: self.num_people = li['numPeople']
            if 'swimmingPool' in li: self.swimming_pool = li['swimmingPool']
            if 'hotTub' in li: self.hot_tub = li['hotTub']
        if 'latitudeLongitude' in js and js['latitudeLongitude']:
            if 'latitude' in js['latitudeLongitude']: self.latitude = js['latitudeLongitude']['latitude']
            if 'longitude' in js['latitudeLongitude']: self.longitude = js['latitudeLongitude']['longitude']

class VueDeviceChannel(object):
    def __init__(self, gid=0, name='', channelNum='1,2,3', channelMultiplier=1.0, channelTypeGid=0):
        self.device_gid: int = gid
        self.name = name
        self.channel_num = channelNum
        self.channel_multiplier = channelMultiplier
        self.channel_type_gid = channelTypeGid
        self.nested_devices = {}

    def from_json_dictionary(self, js: 'dict[str, Any]') -> Self:
        """Populate device channel data from a dictionary extracted from the response json."""
        if 'deviceGid' in js: self.device_gid = js['deviceGid']
        if 'name' in js: self.name = js['name']
        if 'channelNum' in js: self.channel_num = js['channelNum']
        if 'channelMultiplier' in js: self.channel_multiplier = js['channelMultiplier']
        if 'channelTypeGid' in js: self.channel_type_gid = js['channelTypeGid']
        return self
    
    def as_dictionary(self) -> 'dict[str, Any]':
        """Returns a dictionary of the device channel data."""
        return {
            'deviceGid': self.device_gid,
            'name': self.name,
            'channelNum': self.channel_num,
            'channelMultiplier': self.channel_multiplier,
            'channelTypeGid': self.channel_type_gid
        }

class VueUsageDevice(VueDevice):
    def __init__(self, gid=0, timestamp: Optional[datetime.datetime] = None):
        super().__init__(gid=gid)
        self.timestamp = timestamp
        self.channels: dict[str, VueDeviceChannelUsage] = {}

    def from_json_dictionary(self, js: 'dict[str, Any]') -> Self:
        if not js: return self
        if 'deviceGid' in js: self.device_gid = js['deviceGid']
        if 'channelUsages' in js and js['channelUsages']:
            for channel in js['channelUsages']:
                if channel: 
                    populated_channel = VueDeviceChannelUsage(timestamp=self.timestamp).from_json_dictionary(channel)
                    self.channels[populated_channel.channel_num] = populated_channel
        return self

class VueDeviceChannelUsage(VueDeviceChannel):
    def __init__(self, gid: int=0, usage: float=0, channelNum='1,2,3', name='', timestamp: Optional[datetime.datetime] = None):
        super().__init__(gid=gid, name=name, channelNum=channelNum)
        self.name = name
        self.device_gid: int = gid
        self.usage: float = usage
        self.channel_num = channelNum
        self.percentage = 0.0
        self.timestamp = timestamp
        self.nested_devices = {}

    def from_json_dictionary(self, js: 'dict[str, Any]') -> Self:
        """Populate device channel usage data from a dictionary extracted from the response json."""
        if not js: return self
        if 'channelUsages' in js: js = js['channelUsages'] # were given "device" level and we want to work off "channel" level
        if 'name' in js: self.name = js['name']
        if 'deviceGid' in js: self.device_gid = js['deviceGid']
        if 'channelNum' in js: self.channel_num = js['channelNum']
        if 'usage' in js: self.usage = js['usage']
        if 'percentage' in js: self.percentage = js['percentage']
        # Nested device handling
        if 'nestedDevices' in js and js['nestedDevices']:
            for device in js['nestedDevices']:
                if device:
                    populated = VueUsageDevice(timestamp=self.timestamp).from_json_dictionary(device)
                    self.nested_devices[populated.device_gid] = populated
        return self

class OutletDevice(object):
    def __init__(self, gid: int=0, on: bool=False):
        self.device_gid = gid
        self.outlet_on = on
        self.load_gid: int = 0
        self.schedules = []

    def from_json_dictionary(self, js: 'dict[str, Any]') -> Self:
        if 'deviceGid' in js: self.device_gid = js['deviceGid']
        if 'outletOn' in js: self.outlet_on = js['outletOn']
        if 'loadGid' in js: self.load_gid = js['loadGid']
        # don't have support for schedules yet
        return self
    
    def as_dictionary(self) -> 'dict[str, Any]':
        return {
            'deviceGid': self.device_gid,
            'outletOn': self.outlet_on,
            'loadGid': self.load_gid
        }

class ChargerDevice(object):
    def __init__(self, gid: int=0, on: bool=False):
        self.device_gid = gid
        self.charger_on = on
        self.message = ''
        self.status = ''
        self.icon = ''
        self.icon_label = ''
        self.icon_detail_text = ''
        self.fault_text = ''
        self.charging_rate = 0
        self.max_charging_rate = 0
        self.off_peak_schedules_enabled = False
        self.custom_schedules = []
        self.load_gid: int = 0
        self.debug_code = ''
        self.pro_control_code = ''
        self.breaker_pin = ''

    def from_json_dictionary(self, js: 'dict[str, Any]') -> Self:
        if 'deviceGid' in js: self.device_gid = js['deviceGid']
        if 'loadGid' in js: self.load_gid = js['loadGid']
        if 'chargerOn' in js: self.charger_on = js['chargerOn']
        if 'message' in js: self.message = js['message']
        if 'status' in js: self.status = js['status']
        if 'icon' in js: self.icon = js['icon']
        if 'iconLabel' in js: self.icon_label = js['iconLabel']
        if 'iconDetailText' in js: self.icon_detail_text = js['iconDetailText']
        if 'faultText' in js: self.fault_text = js['faultText']
        if 'chargingRate' in js: self.charging_rate = js['chargingRate']
        if 'maxChargingRate' in js: self.max_charging_rate = js['maxChargingRate']
        if 'offPeakSchedulesEnabled' in js: self.off_peak_schedules_enabled = js['offPeakSchedulesEnabled']
        if 'debugCode' in js: self.debug_code = js['debugCode']
        if 'proControlCode' in js: self.pro_control_code = js['proControlCode']
        if 'breakerPIN' in js: self.breaker_pin = js['breakerPIN']
        # don't have support for schedules yet
        return self

    def as_dictionary(self) -> 'dict[str, Any]':
        d = {
            'deviceGid': self.device_gid,
            'loadGid': self.load_gid,
            'chargerOn': self.charger_on,
            'chargingRate': self.charging_rate,
            'maxChargingRate': self.max_charging_rate
        }
        if self.breaker_pin: d['breakerPIN'] = self.breaker_pin
        return d
    
class ChannelType(object):
    def __init__(self, gid: int=0, description: str='', selectable: bool=False) -> None:
        self.channel_type_gid = gid
        self.description = description
        self.selectable = selectable

    def from_json_dictionary(self, js: 'dict[str, Any]') -> Self:
        if 'channelTypeGid' in js: self.channel_type_gid = js['channelTypeGid']
        if 'description' in js: self.description = js['description']
        if 'selectable' in js: self.selectable = js['selectable']
        return self
   
class Vehicle(object):
    def __init__(self, vehicleGid=0, vendor='', apiId='', displayName='', loadGid='', make='', model='', year=0):
        self.vehicle_gid = vehicleGid
        self.vendor = vendor
        self.api_id = apiId
        self.display_name = displayName
        self.loadGid = loadGid
        self.make = make
        self.model = model
        self.year = year

    def from_json_dictionary(self, js):
        if 'vehicleGid' in js: self.vehicle_gid = js['vehicleGid']
        if 'vendor' in js: self.vendor = js['vendor']
        if 'apiId' in js: self.api_id = js['apiId']
        if 'displayName' in js: self.display_name = js['displayName']
        if 'loadGid' in js: self.loadGid = js['loadGid']
        if 'make' in js: self.make = js['make']
        if 'model' in js: self.model = js['model']
        if 'year' in js: self.year = js['year']
        return self

    def as_dictionary(self) -> 'dict[str, Any]':
        return {
            'vehicleGid': self.vehicle_gid,
            'vendor': self.vendor,
            'apiId': self.api_id,
            'displayName': self.displayName,
            'loadGid': self.loadGid,
            'make': self.make,
            'model': self.model,
            'year': self.year
        }

class VehicleStatus(object):
    def __init__(self, vehicleGid=0, vehicleState='', batteryLevel=0, batteryRange=0, chargingState='', chargeLimitPercent = 0, minutesToFullCharge = 0, chargeCurrentRequest = 0, chargeCurrentRequestMax = 0):
        self.vehicle_gid = vehicleGid
        self.vehicle_state = vehicleState
        self.battery_level = batteryLevel
        self.battery_range = batteryRange
        self.charging_state = chargingState
        self.charge_limit_percent = chargeLimitPercent
        self.minutes_to_full_charge = minutesToFullCharge
        self.charge_current_request = chargeCurrentRequest
        self.charge_current_request_max = chargeCurrentRequestMax

    def from_json_dictionary(self, js):

        if 'settings' in js:
            jsv = js['settings']

        if 'vehicleGid' in jsv: self.vehicle_gid = jsv['vehicleGid']
        if 'vehicleState' in jsv: self.vehicle_state = jsv['vehicleState']
        if 'batteryLevel' in jsv: self.battery_level = jsv['batteryLevel']
        if 'batteryRange' in jsv: self.battery_range = jsv['batteryRange']
        if 'chargingState' in jsv: self.charging_state = jsv['chargingState']
        if 'chargeLimitPercent' in jsv: self.charge_limit_percent = jsv['chargeLimitPercent']
        if 'minutesToFullCharge' in jsv: self.minutes_to_full_charge = jsv['minutesToFullCharge']
        if 'chargeCurrentRequest' in jsv: self.charge_current_request = jsv['chargeCurrentRequest']
        if 'chargeCurrentRequestMax' in jsv: self.charge_current_request_max = jsv['chargeCurrentRequestMax']

        return self

    def as_dictionary(self) -> 'dict[str, Any]':
        return {
            'vehicleGid': self.vehicle_gid,
            'vehicleState': self.vehicle_state,
            'batteryLevel': self.battery_level,
            'batteryRange': self.battery_range,
            'chargingState': self.charging_state,
            'chargeLimitPercent': self.charge_limit_percent,
            'minutesToFullCharge': self.minutes_to_full_charge,
            'chargeCurrentRequest': self.charge_current_request,
            'chargeCurrentRequestMax': self.charge_current_request_max
        }



import sys
import datetime
import dateutil

# Our files
from pyemvue.device import VueDevice, VueUsageDevice
from pyemvue.enums import Scale, Unit
from pyemvue.pyemvue import PyEmVue

def print_recursive(usage_dict: dict[int, VueUsageDevice], info: dict[int, VueDevice], scaleBy: float=1, unit='kWh', depth=0):
    for gid, device in usage_dict.items():
        for channelnum, channel in device.channels.items():
            name = channel.name
            if name == 'Main':
                name = info[gid].device_name
            usage = channel.usage or 0
            print('-'*depth, f'{gid} {channelnum} {name} {usage*scaleBy} {unit}')
            if channel.nested_devices:
                print_recursive(channel.nested_devices, info, scaleBy=scaleBy, unit=unit, depth=depth+1)

def main():
    errorMsg = 'Please pass a file containing the "email" and "password" as json.'
    if len(sys.argv) == 1:
        print(errorMsg)
        sys.exit(1)

    filepath = sys.argv[1]
    vue = PyEmVue()
    vue.login(token_storage_file=filepath)
    print('Logged in. Authtoken follows:')
    print(vue.auth.tokens["id_token"])
    print()
    channelTypes = vue.get_channel_types()
    devices = vue.get_devices()
    deviceGids: list[int] = []
    deviceInfo: dict[int, VueDevice] = {}
    for device in devices:
        if not device.device_gid in deviceGids:
            deviceGids.append(device.device_gid)
            deviceInfo[device.device_gid] = device
            print(device.device_gid, device.manufacturer_id, device.model, device.firmware)
            for chan in device.channels:
                channelTypeInfo = next((c for c in channelTypes if c.channel_type_gid == chan.channel_type_gid), None)
                print('\t', chan.device_gid, chan.name, chan.channel_num, chan.channel_multiplier, channelTypeInfo.description if channelTypeInfo else chan.channel_type_gid)
        else:
            deviceInfo[device.device_gid].channels += device.channels
            for chan in device.channels:
                channelTypeInfo = next((c for c in channelTypes if c.channel_type_gid == chan.channel_type_gid), None)
                print('\t', chan.device_gid, chan.name, chan.channel_num, chan.channel_multiplier, channelTypeInfo.description if channelTypeInfo else chan.channel_type_gid)

    monthly, start = vue.get_chart_usage(devices[0].channels[0],scale=Scale.MONTH.value)
    print(monthly[0], 'kwh used since', start.isoformat())
    now = datetime.datetime.now(datetime.timezone.utc)
    midnight=(datetime.datetime
             .now(dateutil.tz.gettz(devices[0].time_zone))
             .replace(hour=0, minute=0, second=0, microsecond=0)
             .astimezone(dateutil.tz.tzutc()))
    yesterday = midnight - datetime.timedelta(days=1)
    yesterday = yesterday.replace(tzinfo=None)
    print('Total usage for today in kwh: ')

    use = vue.get_device_list_usage(deviceGids, now, Scale.DAY.value)
    print_recursive(use, deviceInfo)
    print('Total usage for yesterday in kwh: ')
    for gid, device in deviceInfo.items():
        for chan in device.channels:
            usage = vue.get_chart_usage(chan, yesterday, yesterday+datetime.timedelta(hours=23, minutes=59), Scale.DAY.value)
            if usage and usage[0]:
                print(f'{chan.device_gid} ({chan.channel_num}): {usage[0][0]} kwh')
    print('Average usage over the last minute in watts: ')
    use = vue.get_device_list_usage(deviceGids, None, Scale.MINUTE.value)
    print_recursive(use, deviceInfo, scaleBy=60000, unit='W')

    usage_over_time, start_time = vue.get_chart_usage(devices[0].channels[0], datetime.datetime.now(datetime.timezone.utc)-datetime.timedelta(days=7), datetime.datetime.now(datetime.timezone.utc), scale=Scale.DAY.value, unit=Unit.KWH.value)

    print('Usage for the last seven days starting', start_time.isoformat())
    for usage in usage_over_time:
        print(usage, 'kwh')

    (outlets, chargers) = vue.get_devices_status(devices)
    print('List of Outlets:')
    for outlet in outlets:
        print(f"\t{outlet.device_gid} On? {outlet.outlet_on}")

    print('List of Chargers:')
    for charger in chargers:
        print(f"\t{charger.device_gid} On? {charger.charger_on} Charge rate: {charger.charging_rate}/{charger.max_charging_rate} Status: {charger.status}")

    try:
        vehicles = vue.get_vehicles()
        print('List of Vehicles')
        for vehicle in vehicles:
            print(f'\t{vehicle.vehicle_gid} ({vehicle.display_name}) - {vehicle.year} {vehicle.make} {vehicle.model}')

        print('List of Vehicle Statuses')
        for vehicle in vehicles:
            vehicleStatus = vue.get_vehicle_status(vehicle.vehicle_gid)
            if vehicleStatus:
                print(f'\t{vehicleStatus.vehicle_gid} {vehicleStatus.vehicle_state} - Charging: {vehicleStatus.charging_state} Battery level: {vehicleStatus.battery_level}')
            else:
                print(f'\t{vehicle.vehicle_gid} - No status available')
    except Exception as e:
        print(f'Error getting vehicles: {e}')

if __name__ == '__main__':
    main()




# This script parses out unique API calls from the monitored traffic and provides example bodies
# Eventually it will likely support identifying new/removed/changed calls and content
import os
import json
from urllib.parse import urlparse

root = ''

unique_urls = {}

for path, dirs, files in os.walk(root):
    for file in files:
        if file == 'request.json':
            with open(os.path.join(path, file)) as f:
                request = json.load(f)
            method = request['method']
            url = urlparse(request['url'])
            key = f'{method} {url.path}'
            if url.path == '/AppAPI':
                # this is actually a bunch of methods
                apiMethod = url.query.split('&', 1)[0]
                key = f'{key}/{apiMethod}'
            if key not in unique_urls:
                
                sample = {
                    'url': url.geturl()
                }
                if 'request_body.txt' in files:
                    with open(os.path.join(path, 'request_body.txt')) as b:
                        try:
                            sample['body'] = json.load(b)
                        except:
                            sample['body'] = b.read()
                if 'response_body.json' in files:
                    with open(os.path.join(path, 'response_body.json')) as b:
                        sample['response'] = json.load(b)
                unique_urls[key] = sample

# dump = json.dumps(unique_urls, indent=2)
with open('api_data.json', 'w') as f:
    json.dump(unique_urls, f, indent=2)


# SimulatorState of the Emporia API. Holds all the state needed to simulate the API.

import json
import os

from simulator.models import *


class SimulatorState(object):
    def __init__(self):
        self.customer = SimulatorCustomer(
            customerGid=1234,
            email="simulator@simulator.sim",
            firstName="Sim",
            lastName="Ulator",
            createdAt=datetime.datetime.utcnow(),
        )
        # Hold the usage for each channel. deviceGid_channelNum : usage
        self.usage_dict_1min: dict[str, Optional[float]] = {}
        self.devices: list[SimulatorDevice] = []
        self.channel_types: list[ChannelType] = []
        self.outlets: list[SimulatorOutlet] = []
        self.chargers: list[SimulatorCharger] = []
        self.default_location_props = SimulatorLocationProperties(
            timeZone="America/New_York",
            latitudeLongitude=SimulatorLatitudeLongitude(
                latitude=40.0, longitude=-74.0
            ),
            utilityRateGid=None,
            deviceName=None,
            deviceGid=1234,
            zipCode="10001",
            billingCycleStartDay=1,
            usageCentPerKwHour=15.0,
            peakDemandDollarPerKw=0.0,
            locationInformation=SimulatorLocationInformation(
                airConditioning="true",
                heatSource="electricFurnace",
                locationSqFt="2000",
                numElectricCars="0",
                locationType="houseMultiLevel",
                numPeople="4",
            ),
        )

        script_dir = os.path.dirname(__file__)
        channel_types_file = script_dir + "/channel_types.json"
        self.load_channel_types(channel_types_file)

    def load_channel_types(self, channel_types_file: str):
        with open(channel_types_file, "r") as f:
            self.channel_types = json.load(f)

    def get_customers_devices(self) -> CustomerDevicesResponse:
        return CustomerDevicesResponse(
            customerGid=self.customer.customerGid,
            email=self.customer.email,
            firstName=self.customer.firstName,
            lastName=self.customer.lastName,
            createdAt=self.customer.createdAt,
            devices=self.devices,
        )

    def get_status(self) -> StatusResponse:
        return StatusResponse(
            devicesConnected=[device.deviceConnected for device in self.devices],
            evChargers=self.chargers,
            loads=[],
            batteries=[],
            outlets=self.outlets,
        )

    # state.get_devices_usage(deviceGids, instant, scale, energyUnit)
    def get_devices_usage(
        self,
        deviceGids: Optional[str],
        instant: datetime.datetime,
        scale: str,
        energyUnit: str,
    ) -> DeviceUsageResponse:
        if scale != "1MIN" or energyUnit != "KilowattHours":
            raise Exception(
                f"Scale {scale} or energyUnit {energyUnit} not yet supported"
            )
        deviceListUsage = DeviceListUsage(
            energyUnit=energyUnit, scale=scale, instant=instant, devices=[]
        )
        response = DeviceUsageResponse(deviceListUsages=deviceListUsage)
        # build the usage tree by recursively calling build_tree
        root = self.build_tree(None, None)
        deviceListUsage.devices = root
        return response

    def build_tree(
        self, parentDeviceGid: Optional[int], parentChannelNum: Optional[str]
    ) -> list[DeviceUsage]:
        devices_at_level: list[DeviceUsage] = []
        # We end up looping through all of the devices each time, but that's ok for now
        # we can optimize later if it becomes a performance concern
        for device in self.devices:
            if (
                device.parentDeviceGid == parentDeviceGid
                and device.parentChannelNum == parentChannelNum
            ):
                # handle the main device, then any sub-devices
                device_usage = DeviceUsage(deviceGid=device.deviceGid, channelUsages=[])
                channel_usages: list[ChannelUsage] = []
                for channel in device.channels:
                    # get the usage for this channel
                    usage = self.usage_dict_1min.get(
                        f"{device.deviceGid}_{channel.channelNum}", 0.0
                    )
                    channel_usages.append(
                        ChannelUsage(
                            name=channel.name or "Main",
                            percentage=0.0,
                            usage=usage,
                            deviceGid=device.deviceGid,
                            channelNum=channel.channelNum,
                            nestedDevices=self.build_tree(
                                device.deviceGid, channel.channelNum
                            ),
                        )
                    )
                for sub_device in device.devices:
                    for channel in sub_device.channels:
                        # get the usage for this channel
                        usage = self.usage_dict_1min.get(
                            f"{sub_device.deviceGid}_{channel.channelNum}", 0.0
                        )
                        channel_usages.append(
                            ChannelUsage(
                                name=channel.name or "Main",
                                percentage=0.0,
                                usage=usage,
                                deviceGid=sub_device.deviceGid,
                                channelNum=channel.channelNum,
                                nestedDevices=self.build_tree(
                                    sub_device.deviceGid, channel.channelNum
                                ),
                            )
                        )
                # There is a special Balance channel that appears when there are multiple channels on a device
                if len(channel_usages) > 1:
                    balance_usage = self.usage_dict_1min.get(
                        f"{device.deviceGid}_Balance", 0.0
                    )
                    channel_usages.append(
                        ChannelUsage(
                            name="Balance",
                            percentage=0.0,
                            usage=balance_usage,
                            deviceGid=device.deviceGid,
                            channelNum="Balance",
                            nestedDevices=[],
                        )
                    )
                device_usage.channelUsages = channel_usages
                devices_at_level.append(device_usage)
        return devices_at_level

    def set_location_properties(
        self, location_properties: SimulatorLocationProperties, propagate: bool = True
    ):
        self.default_location_props = location_properties.model_copy()
        if propagate:
            for device in self.devices:
                # get the device name before we overwrite it
                device_name = device.locationProperties.deviceName
                # overwrite the location properties
                device.locationProperties = location_properties.model_copy()
                # fix the device name and gid
                device.locationProperties.deviceGid = device.deviceGid
                device.locationProperties.deviceName = device_name

    def add_outlet(
        self,
        gid: int,
        name: Optional[str],
        on: bool = True,
        parentDeviceGid: Optional[int] = None,
        parentChannelNum: Optional[str] = None,
    ) -> SimulatorOutlet:
        # if any devices already have this gid, throw an error
        for device in self.devices:
            if device.deviceGid == gid:
                raise Exception(f"Device gid {gid} already exists")
        # Create a basic outlet
        outlet = SimulatorOutlet(deviceGid=gid, outletOn=on, loadGid=0)
        self.outlets.append(outlet)
        # Create a full device for the outlet
        outlet_device = SimulatorDevice(
            deviceGid=gid,
            manufacturerDeviceId="B0000C89109103c1298",
            model="SSO001",
            firmware="Outlet-488",
            parentDeviceGid=parentDeviceGid,
            parentChannelNum=parentChannelNum,
            locationProperties=self.default_location_props.model_copy(),
            outlet=outlet,
            evCharger=None,
            battery=None,
            deviceConnected=SimulatorDeviceConnected(
                deviceGid=gid, connected=True, offlineSince=None
            ),
            devices=[],
            channels=[
                SimulatorChannel(
                    deviceGid=gid,
                    name=None,
                    channelNum="1,2,3",
                    channelMultiplier=1.0,
                    channelTypeGid=23,
                )
            ],
        )

        outlet_device.locationProperties.deviceGid = gid
        outlet_device.locationProperties.deviceName = name
        self.devices.append(outlet_device)
        return outlet

    def add_charger(
        self,
        gid: int,
        name: Optional[str],
        on: bool = True,
        breakerSize: int = 50,
        parentDeviceGid: Optional[int] = None,
        parentChannelNum: Optional[str] = None,
    ) -> SimulatorCharger:
        # if any devices already have this gid, throw an error
        for device in self.devices:
            if device.deviceGid == gid:
                raise Exception(f"Device gid {gid} already exists")
        # Create a basic charger
        charger = SimulatorCharger(
            deviceGid=gid,
            loadGid=0,
            chargerOn=on,
            chargingRate=4 * breakerSize // 5,
            maxChargingRate=4 * breakerSize // 5,
        )
        self.chargers.append(charger)
        # Create a full device for the charger
        charger_device = SimulatorDevice(
            deviceGid=gid,
            manufacturerDeviceId="D2112A04B4AC6C4523421",
            model="VVDN01",
            firmware="EVCharger-467",
            parentDeviceGid=parentDeviceGid,
            parentChannelNum=parentChannelNum,
            locationProperties=self.default_location_props.model_copy(),
            outlet=None,
            evCharger=charger,
            battery=None,
            deviceConnected=SimulatorDeviceConnected(
                deviceGid=gid, connected=True, offlineSince=None
            ),
            devices=[],
            channels=[
                SimulatorChannel(
                    deviceGid=gid,
                    name=None,
                    channelNum="1,2,3",
                    channelMultiplier=1.0,
                    channelTypeGid=25,
                )
            ],
        )

        charger_device.locationProperties.deviceGid = gid
        charger_device.locationProperties.deviceName = name
        self.devices.append(charger_device)
        return charger

    def add_vue(
        self,
        gid: int,
        name: Optional[str],
        channelCount: int = 8,
        parentDeviceGid: Optional[int] = None,
        parentChannelNum: Optional[str] = None,
    ) -> SimulatorDevice:
        # if any devices already have this gid, throw an error
        for device in self.devices:
            if device.deviceGid == gid:
                raise Exception(f"Device gid {gid} already exists")
        # Create a basic vue
        vue = SimulatorDevice(
            deviceGid=gid,
            manufacturerDeviceId="A2112A04B4AC67B2B548201",
            model="VUE02",
            firmware="Vue2-434",
            parentDeviceGid=parentDeviceGid,
            parentChannelNum=parentChannelNum,
            locationProperties=self.default_location_props.model_copy(),
            outlet=None,
            evCharger=None,
            battery=None,
            deviceConnected=SimulatorDeviceConnected(
                deviceGid=gid, connected=True, offlineSince=None
            ),
            devices=[],
            channels=[],
        )
        # Create the main channel
        vue.channels.append(
            SimulatorChannel(
                deviceGid=gid,
                name=None,
                channelNum="1,2,3",
                channelMultiplier=1.0,
                channelTypeGid=None,
            )
        )
        # Create the other channels as children of the "WAT001" device
        if channelCount > 0:
            subDevice = SimulatorSubDevice(
                deviceGid=gid,
                manufacturerDeviceId="SXA2112A04B4AC67B2F12345",
                model="WAT001",
                firmware=None,
                channels=[],
            )
            vue.devices.append(subDevice)
            for i in range(1, channelCount + 1):
                subDevice.channels.append(
                    SimulatorChannel(
                        deviceGid=gid,
                        name="Channel " + str(i),
                        channelNum=str(i),
                        channelMultiplier=1.0,
                        channelTypeGid=18,
                    )
                )
        # Add the device to the list
        vue.locationProperties.deviceGid = gid
        vue.locationProperties.deviceName = name
        self.devices.append(vue)
        return vue

    def delete_device(self, gid: int) -> Optional[SimulatorDevice]:
        for device in self.devices:
            if device.deviceGid == gid:
                self.devices.remove(device)
                return device
        return None

    def set_channel_1min_watts(self, deviceGid: int, channelNum: str, watts: Optional[float]):
        # convert from watts to kilowatt hours used over a 1 minute period
        if watts is None:
            usage = None
        else:
            scaler = 60 * 1000  # 60 minutes/hr * 1000 watts/kilowatt
            usage = watts / scaler
        self.set_channel_1min_usage(deviceGid, channelNum, usage)

    def set_channel_1min_usage(self, deviceGid: int, channelNum: str, usage: Optional[float]):
        self.usage_dict_1min[f"{deviceGid}_{channelNum}"] = usage



import time
from typing import Any, Optional, Union
import requests
import datetime
import json
from dateutil.parser import parse

# Our files
from pyemvue.auth import Auth, SimulatedAuth
from pyemvue.enums import Scale, Unit
from pyemvue.customer import Customer
from pyemvue.device import ChargerDevice, VueDevice, OutletDevice, VueDeviceChannel, VueDeviceChannelUsage, VueUsageDevice, ChannelType, Vehicle, VehicleStatus

API_ROOT = 'https://api.emporiaenergy.com'
API_CHANNELS = 'devices/{deviceGid}/channels'
API_CHANNEL_TYPES = 'devices/channels/channeltypes'
API_CHARGER = 'devices/evcharger'
API_CHART_USAGE = 'AppAPI?apiMethod=getChartUsage&deviceGid={deviceGid}&channel={channel}&start={start}&end={end}&scale={scale}&energyUnit={unit}'
API_CUSTOMER = 'customers'
API_CUSTOMER_DEVICES = 'customers/devices'
API_DEVICES_USAGE = 'AppAPI?apiMethod=getDeviceListUsages&deviceGids={deviceGids}&instant={instant}&scale={scale}&energyUnit={unit}'
API_DEVICE_PROPERTIES = 'devices/{deviceGid}/locationProperties'
API_GET_STATUS = 'customers/devices/status'
API_OUTLET = 'devices/outlet'
API_VEHICLES = 'customers/vehicles'
API_VEHICLE_STATUS = 'vehicles/v2/settings?vehicleGid={vehicleGid}'

API_MAINTENANCE = 'https://s3.amazonaws.com/com.emporiaenergy.manual.ota/maintenance/maintenance.json'

class PyEmVue(object):
    def __init__(self, connect_timeout: float = 6.03, read_timeout: float = 10.03):
        self.username = None
        self.token_storage_file = None
        self.customer = None
        self.connect_timeout = connect_timeout
        self.read_timeout = read_timeout

    def down_for_maintenance(self) -> Optional[str]:
        """Checks to see if the API is down for maintenance, returns the reported message if present."""
        response = requests.get(API_MAINTENANCE)
        if response.status_code == 404: return None
        if response.text:
            j = response.json()
            if 'msg' in j:
                return j['msg']

    def get_devices(self) -> 'list[VueDevice]':
        """Get all devices under the current customer account."""
        response = self.auth.request('get', API_CUSTOMER_DEVICES)
        response.raise_for_status()
        devices: list[VueDevice] = []
        if response.text:
            j = response.json()
            if 'devices' in j:
                for dev in j['devices']:
                    devices.append(VueDevice().from_json_dictionary(dev))
                    if 'devices' in dev:
                        for subdev in dev['devices']:
                            devices.append(VueDevice().from_json_dictionary(subdev))
        return devices

    def populate_device_properties(self, device: VueDevice) -> VueDevice:
        """Get details about a specific device"""
        url = API_DEVICE_PROPERTIES.format(deviceGid=device.device_gid)
        response = self.auth.request('get', url)
        response.raise_for_status()
        if response.text:
            j = response.json()
            device.populate_location_properties_from_json(j)
        return device

    def update_channel(self, channel: VueDeviceChannel) -> VueDeviceChannel:
        """Update the channel with the provided state."""
        url = API_CHANNELS.format(deviceGid=channel.device_gid)
        response = self.auth.request('put', url, json=channel.as_dictionary())
        response.raise_for_status()
        if response.text:
            j = response.json()
            channel.from_json_dictionary(j)
        return channel

    def get_customer_details(self) -> Optional[Customer]:
        """Get details for the current customer."""
        response = self.auth.request('get', API_CUSTOMER)
        response.raise_for_status()
        if response.text:
            j = response.json()
            return Customer().from_json_dictionary(j)
        return None


    def get_device_list_usage(self, deviceGids: Union[str, 'list[str]'], instant: Optional[datetime.datetime], scale=Scale.SECOND.value, unit=Unit.KWH.value) -> 'dict[int, VueUsageDevice]':
        """Returns a nested dictionary of VueUsageDevice and VueDeviceChannelUsage with the total usage of the devices over the specified scale. Note that you may need to scale this to get a rate (1MIN in kw = 60*result)"""
        if not instant: instant = datetime.datetime.now(datetime.timezone.utc)
        gids = deviceGids
        if isinstance(deviceGids, list):
            gids = '+'.join(map(str, deviceGids))

        url = API_DEVICES_USAGE.format(deviceGids=gids, instant=_format_time(instant), scale=scale, unit=unit)
        max_attempts = 3
        attempts = 0
        success = False

        while attempts < max_attempts and not success:
            if attempts > 0:
                # if we're retrying, wait a bit before trying again
                delay = 2**attempts
                time.sleep(delay)
            attempts += 1
            response = self.auth.request('get', url)
            devices: dict[int, VueUsageDevice] = {}
            if response.status_code == 200 and response.text:
                j = response.json()
                if 'deviceListUsages' in j and 'devices' in j['deviceListUsages']:
                    timestamp = parse(j['deviceListUsages']['instant'])
                    for device in j['deviceListUsages']['devices']:
                        populated = VueUsageDevice(timestamp=timestamp).from_json_dictionary(device)
                        devices[populated.device_gid] = populated
                        for _, channel_usage in populated.channels.items():
                            if channel_usage.usage is not None:
                                success = True
                            else:
                                # if the usage is None then we have bad data. Retry from the top.
                                success = False
                                break
                        if not success:
                            break
                else:
                    success = False
            else:
                success = False
        if response:
            response.raise_for_status()
        return devices

    def get_chart_usage(self, channel: Union[VueDeviceChannel, VueDeviceChannelUsage], start: Optional[datetime.datetime] = None, end: Optional[datetime.datetime] = None, scale=Scale.SECOND.value, unit=Unit.KWH.value) -> 'tuple[list[float], Optional[datetime.datetime]]':
        """Gets the usage over a given time period and the start of the measurement period. Note that you may need to scale this to get a rate (1MIN in kw = 60*result)"""
        if channel.channel_num in ['MainsFromGrid', 'MainsToGrid']:
            # These is not populated for the special Mains data as of right now
            return [], start
        if not start: start = datetime.datetime.now(datetime.timezone.utc)
        if not end: end = datetime.datetime.now(datetime.timezone.utc)
        url = API_CHART_USAGE.format(deviceGid=channel.device_gid, channel=channel.channel_num, start=_format_time(start), end=_format_time(end), scale=scale, unit=unit)
        response = self.auth.request('get', url)
        response.raise_for_status()
        usage: list[float] = []
        instant = start
        if response.text:
            j = response.json()
            if 'firstUsageInstant' in j: instant = parse(j['firstUsageInstant'])
            if 'usageList' in j: usage = j['usageList']
        return usage, instant

    def get_outlets(self) -> 'list[OutletDevice]':
        """ Return a list of outlets linked to the account. Deprecated, use get_devices_status instead."""
        response = self.auth.request('get', API_GET_STATUS)
        response.raise_for_status()
        outlets = []
        if response.text:
            j = response.json()
            if j and 'outlets' in j and j['outlets']:
                for raw_outlet in j['outlets']:
                    outlets.append(OutletDevice().from_json_dictionary(raw_outlet))
        return outlets

    def update_outlet(self, outlet: OutletDevice, on: Optional[bool]=None) -> OutletDevice:
        """ Primarily to turn an outlet on or off. If the on parameter is not provided then uses the value in the outlet object.
            If on parameter provided uses the provided value."""
        if on is not None:
            outlet.outlet_on = on

        response = self.auth.request('put', API_OUTLET, json=outlet.as_dictionary())
        response.raise_for_status()
        outlet.from_json_dictionary(response.json())
        return outlet

    def get_chargers(self) -> 'list[ChargerDevice]':
        """ Return a list of EVSEs/chargers linked to the account. Deprecated, use get_devices_status instead."""
        response = self.auth.request('get', API_GET_STATUS)
        response.raise_for_status()
        chargers = []
        if response.text:
            j = response.json()
            if j and 'evChargers' in j and j['evChargers']:
                for raw_charger in j['evChargers']:
                    chargers.append(ChargerDevice().from_json_dictionary(raw_charger))
        return chargers

    def update_charger(self, charger: ChargerDevice, on: Optional[bool] = None, charge_rate: Optional[int] = None) -> ChargerDevice:
        """ Primarily to enable/disable an evse/charger. The on and charge_rate parameters override the values in the object if provided"""
        if on is not None:
            charger.charger_on = on
        if charge_rate:
            charger.charging_rate = charge_rate

        response = self.auth.request('put', API_CHARGER, json=charger.as_dictionary())
        response.raise_for_status()
        charger.from_json_dictionary(response.json())
        return charger

    def get_devices_status(self, device_list: Optional['list[VueDevice]'] = None) -> 'tuple[list[OutletDevice], list[ChargerDevice]]':
        """Gets the list of outlets and chargers. If device list is provided, updates the connected status on each device."""
        response = self.auth.request('get', API_GET_STATUS)
        response.raise_for_status()
        chargers: list[ChargerDevice] = []
        outlets: list[OutletDevice] = []
        if response.text:
            j = response.json()
            if j and 'evChargers' in j and j['evChargers']:
                for raw_charger in j['evChargers']:
                    chargers.append(ChargerDevice().from_json_dictionary(raw_charger))
            if j and 'outlets' in j and j['outlets']:
                for raw_outlet in j['outlets']:
                    outlets.append(OutletDevice().from_json_dictionary(raw_outlet))
            if device_list and j and 'devicesConnected' in j and j['devicesConnected']:
                for raw_device_data in j['devicesConnected']:
                    if raw_device_data and 'deviceGid' in raw_device_data and raw_device_data['deviceGid']:
                        for device in device_list:
                            if device.device_gid == raw_device_data['deviceGid']:
                                device.connected = raw_device_data['connected']
                                device.offline_since = raw_device_data['offlineSince']
                                break

        return (outlets, chargers)
    
    def get_channel_types(self) -> 'list[ChannelType]':
        """Gets the list of channel types"""
        response = self.auth.request('get', API_CHANNEL_TYPES)
        response.raise_for_status()
        channel_types: list[ChannelType] = []
        if response.text:
            j = response.json()
            if j:
                for raw_channel_type in j:
                    channel_types.append(ChannelType().from_json_dictionary(raw_channel_type))
        return channel_types

    def get_vehicles(self) -> 'list[Vehicle]':
        """Get all vehicles under the current customer account."""
        response = self.auth.request('get', API_VEHICLES)
        response.raise_for_status()
        vehicles: list[Vehicle] = []
        if response.text:
            j = response.json()
            for veh in j:
                vehicles.append(Vehicle().from_json_dictionary(veh))
        return vehicles

    def get_vehicle_status(self, vehicle_gid: int) -> Optional[VehicleStatus]:
        """Get details for the current vehicle."""
        url = API_VEHICLE_STATUS.format(vehicleGid=vehicle_gid)
        response = self.auth.request('get', url)
        response.raise_for_status()
        if response.text:
            j = response.json()
            return VehicleStatus().from_json_dictionary(j)
        return None

    def login(self, username: Optional[str]=None, password: Optional[str]=None, id_token: Optional[str]=None, access_token: Optional[str]=None, refresh_token: Optional[str]=None, token_storage_file: Optional[str]=None) -> bool:
        """ Authenticates the current user using access tokens if provided or username/password if no tokens available.
            Provide a path for storing the token data that can be used to reauthenticate without providing the password.
            Tokens stored in the file are updated when they expire.
        """
        # try to pull data out of the token storage file if present
        self.username = username.lower() if username else None
        if token_storage_file: self.token_storage_file = token_storage_file
        if not password and not id_token and token_storage_file:
            with open(token_storage_file, 'r') as f:
                data = json.load(f)
                if 'id_token' in data: id_token = data['id_token']
                if 'access_token' in data: access_token = data['access_token']
                if 'refresh_token' in data: refresh_token = data['refresh_token']
                if 'username' in data: self.username = data['username']
                if 'password' in data: password = data['password']

        self.auth = Auth(
            host=API_ROOT,
            username=self.username,
            password=password,
            connect_timeout=self.connect_timeout,
            read_timeout=self.read_timeout,
            tokens={
                'access_token': access_token,
                'id_token': id_token,
                'refresh_token': refresh_token
            },
            token_updater=self._store_tokens
        )

        if self.auth.tokens:
            self.username = self.auth.get_username()
            self.customer = self.get_customer_details()
            self._store_tokens(self.auth.tokens)
        return self.customer is not None
    
    def login_simulator(self, host: str, username: Optional[str]=None, password: Optional[str]=None) -> bool:
        self.username = username.lower() if username else None
        self.auth = SimulatedAuth(host=host, username=self.username, password=password)
        self.customer = self.get_customer_details()
        return self.customer is not None

    def _store_tokens(self, tokens: 'dict[str, Any]'):
        if not self.token_storage_file: return
        if self.username:
            tokens['username'] = self.username
        with open(self.token_storage_file, 'w') as f:
            json.dump(tokens, f, indent=2)

def _format_time(time: datetime.datetime) -> str:
    '''Convert time to utc, then format'''
    # check if aware
    if time.tzinfo and time.tzinfo.utcoffset(time) is not None:
        # aware, convert to utc
        time = time.astimezone(datetime.timezone.utc)
    else:
        #unaware, assume it's already utc
        time = time.replace(tzinfo=
        datetime.timezone.utc)
    time = time.replace(tzinfo=None) # make it unaware
    return time.isoformat()+'Z'



from pyemvue.pyemvue import PyEmVue

vue = PyEmVue()
vue.login(token_storage_file='keys.json')
print('Logged in.')
print()

devices = vue.get_devices()
outlets, chargers = vue.get_devices_status()
for device in devices:
    print(device.device_gid, device.manufacturer_id, device.model, device.firmware)
    if device.outlet:
        print('Found an outlet! On?', device.outlet.outlet_on)
    if device.ev_charger:
        print(f'Found an EV Charger! On? {device.ev_charger.charger_on} Charge rate: {device.ev_charger.charging_rate}A/{device.ev_charger.max_charging_rate}A')
    for chan in device.channels:
        print('\t', chan.device_gid, chan.name, chan.channel_num, chan.channel_multiplier)
    
if len(outlets) > 0:
    print("Discovered {} outlets. Press enter to turn them all off.".format(len(outlets)))
    input()
    for outlet in outlets:
        print("Turning off {}".format(outlet.device_gid))
        vue.update_outlet(outlet, False)
    print("Outlets turned off. Press enter to turn them on.")
    input()
    for outlet in outlets:
        print("Turning on {}".format(outlet.device_gid))
        vue.update_outlet(outlet, True)
else:
    print("No outlets discovered.")
